// z80snap.cpp
// Revision 5-oct-2003
// (C) 2003 Julian Albo

#include "z80snap.h"

#include <math.h>
// Don't use cmath for compatibility with old compilers.

// For debugging
#include <iostream>
using std::cerr;
using std::endl;

namespace {

struct BlockHead {
	std::streamsize length;
	int page;
	bool read (std::istream & is);
};

bool BlockHead::read (std::istream & is)
{
	z80snap::CHAR rawbh [3];
	is.read (reinterpret_cast <char *> (rawbh), 3);
	if (is.eof () )
		return false;
	if (! is || is.gcount () != 3)
		throw "Error reading block header";
	length= rawbh [0] | rawbh [1] << 8;
	page= rawbh [2];
	return true;
}

} // namespace

z80snap::Header::Header (std::istream & is, bool debug)
{
	const std::streamsize headsize= 30;
	CHAR rawhead [headsize];
	is.read (reinterpret_cast <char *> (rawhead), headsize);
	if (! is || std::streamsize (is.gcount () ) != headsize)
		throw ErrorReadingHeader ();

	regs.A= rawhead [0];
	regs.F= rawhead [1];
	regs.BC= rawhead [2] | rawhead [3] << 8;
	regs.HL= rawhead [4] | rawhead [5] << 8;
	regs.PC= rawhead [6] | rawhead [7] << 8;
	regs.SP= rawhead [8] | rawhead [9] << 8;

	CHAR value12= rawhead [12];
	if (value12 == 255)
		value12= 1;
	compressed= value12 & 0xA0; // Bit 5

	regs.DE= rawhead [13] | rawhead [14] << 8;
	regs.BCp= rawhead [15] | rawhead [16] << 8;
	regs.DEp= rawhead [17] | rawhead [18] << 8;
	regs.HLp= rawhead [19] | rawhead [20] << 8;
	regs.Ap= rawhead [21];
	regs.Fp= rawhead [22];
	regs.IY= rawhead [23] | rawhead [24] << 8;
	regs.IX= rawhead [25] | rawhead [26] << 8;

	oldstyle= regs.PC != 0;
	if (! oldstyle)
	{
		// Extended format

		// Get extension size.

		CHAR rawlength [2];
		is.read (reinterpret_cast <char *> (rawlength), 2);
		if (! is || is.gcount () != 2)
			throw ErrorReadingHeader ();
		std::streamsize extlength=
			rawlength [0] | rawlength [1] << 8;
		if (debug)
			cerr << "Extension lenght: " << extlength << endl;
		if (extlength == 0)
			throw ErrorReadingHeader ();

		// Load header extension.

		CHAR * rawext= new CHAR [extlength];
		is.read (reinterpret_cast <char *> (rawext), extlength);
		if (! is || std::streamsize (is.gcount () ) != extlength)
		{
			delete [] rawext;
			throw ErrorReadingHeader ();
		}
		const std::streamsize extoff= headsize + 2;
		regs.PC= rawext [32 - extoff] | rawext [33 - extoff];

		if (debug)
			cerr << "Byte type: " <<
				int (rawext [34 - extoff] ) << endl;
		switch (rawext [34 - extoff] )
		{
		case 0:
		case 1:
		case 2: // Added type 2 according to libspectrum.
			// Spectrum 48K
			type= Type48;
			break;
		case 3:
			if (extlength == 23)
				type= Type128;
			else
				type= Type48;
			break;
		case 4:
		case 5:
		case 6:
			// Spectrum 128K
			type= Type128;
			break;
		case 7:
		case 8: // Added type 8 according to libspectrum
			// Spectrum +3
			type= Type128;
			break;
		case 12:
			// Spectrum +2
			type= Type128;
			break;
		case 13:
			// Spectrum +2A
			type= Type128;
			break;
		default:
			type= TypeUnknown;
		}

		// Spectaculator extension: if bit 7 of byte 37
		// is on the values of hardware are changed.
		if (rawext [37 - extoff] & 0x80)
		{
			if (type == Type48)
				type= Type16;
		}

		delete [] rawext;

	}
	else
	{
		// Old format

		type= Type48;
	}

}

z80snap::Snapshot::Snapshot (std::istream & is, bool debug) :
	head (is)
{
	CHAR discard [16 * 1024];
	if (head.isoldstyle () )
	{
		if (head.iscompressed () )
			loadblockv1 (is, debug);
		else
		{
			if (debug)
				cerr << "Loading V1 uncompressed block" <<
					endl;
			is.read (reinterpret_cast <char *> (memory),
				48 * 1024);
			std::streamsize r= is.gcount ();
			if (r != 16 * 1024 && r != 48 * 1024)
				std::cerr << "WARNING: uncompressed block "
					" of size " << r << endl;
		}
	}
	else
	{
		// Read data blocks.

		MachineType type= head.gettype ();
		BlockHead block;
		bool none= true;
		while (block.read (is) )
		{
			none= false;
			if (debug)
			{
				cerr << "Block length: " <<
					block.length << endl <<
					"Page: " << block.page << endl;
			}
			CHAR * aux;
			switch (block.page)
			{
			case 3:
				if (type == Type48)
					aux= discard;
				else
					aux= memory + 0xC000; // Page 0
				break;
			case 4:
				if (type == Type48)
					aux= memory + 0x8000;
				else
					aux= discard; // Page 1
				break;
			case 5:
				if (type == Type48)
					aux= memory + 0xC000;
				else
					aux= memory + 0x8000; // Page 2
				break;
			case 8:
				aux= memory + 0x4000; // Page 5 on 128
				break;
			default:
				aux= discard;
			}
			loadblock (is, aux, block.length, debug);
		}
		if (none)
			throw "There is no blocks!";
	}
}

namespace {

void check (size_t n)
{
	if (n > 0xFFFF)
		throw "PEEK out of address space";
}

}

z80snap::CHAR z80snap::Snapshot::peek (size_t n)
{
	check (n);
	return memory [n];
}

size_t z80snap::Snapshot::peek16 (size_t n)
{
	check (n);
	return memory [n] | memory [n + 1] << 8;
}

size_t z80snap::Snapshot::peeklinenumber (size_t n)
{
	check (n);
	return memory [n]  << 8 | memory [n + 1];
}

double z80snap::Snapshot::peeknumber (size_t n)
{
	// Returns a number from the Spectrum Basic numeric format.
	CHAR b0= memory [n];
	CHAR b1= memory [n + 1];
	CHAR b2= memory [n + 2];
	CHAR b3= memory [n + 3];
	CHAR b4= memory [n + 4];

	// Special case for integers between -65535 and 65535.
	// The case when b0 and b4 are 0, b1 is 255 and b2 and b3 are 0
	// seem special. When POKEing it in the Spectrum it reports the
	// value -1E-38 but interpreted here as floating gives
	// -2.92726E-39. As last resource, I return -1E-38 directly to
	// be consistent with the Spectrum. I think that combination
	// can't be the result of a normal calculation.

	if (b0 == 0 && b4 == 0 && (b1 == 0 || b1 == 255) )
	{
		size_t n= b2 | b3 << 8;
		if (b1)
		{
			if (n)
				return -static_cast <double> (65536 - n);
			else return -1e-38;
		}
		return static_cast <double> (n);
	}

	// First bit of mantissa is always 1.
	size_t mantissa= ( (b1 & 0x7F) | 0x80) << 24 |
		b2 << 16 | b3 << 8 | b4;
	// First bit of b1 is the sign.
	bool negative= b1 & 0x80;
	// Exponent - 128, and 32 bits after the point.
	int exponent= b0 - 160;
	double result= 	pow (2, exponent) * mantissa;
	if (negative)
		result= -result;
	return result;
}

void z80snap::Snapshot::loadblock (std::istream & is,
	CHAR * ptr, std::streamsize length, bool debug)
{
	if (length == 0xFFFF)
	{
		is.read (reinterpret_cast <char *> (ptr), 16 * 1024);
		if (! is || is.gcount () != 16 * 1024)
			throw "Error reading block header";
	}
	else
	{
		size_t pos= 0;
		const CHAR mark= '\xED';
		CHAR c;
		while (length > 0)
		{
			if (pos >= 16 * 1024)
				throw "Block too long";
			c= is.get ();
			if (! is)
				throw "Error reading block";
			--length;
			if (c != mark)
			{
				* (ptr++)= c;
				++pos;
			}
			else
			{
				c= is.get ();
				if (! is)
				{
					* (ptr++)= mark;
					++pos;
					continue;
				}
				--length;
				if (c != mark)
				{
					* (ptr++)= mark;
					++pos;
					if (pos >= 16 * 1024)
						throw "Block too long";
					* (ptr++)= c;
					++pos;
					continue;
				}
				CHAR times= is.get ();
				if (! is)
					throw "Error reading block";
				--length;
				c= is.get ();
				if (! is)
					throw "Error reading block";
				--length;
				if (debug)
					cerr << "Byte " << int (c) <<
						" repeated " <<
						int (times) << endl;
				for ( ; times > 0; --times)
				{
					if (pos >= 16 * 1024)
						throw "Block too long";
					* (ptr++)= c;
					++pos;
				}
			}
		}
		if (pos != 16 * 1024)
			cerr << "WARNING: uncompressed block is not 16K long"
				<< endl;
	}
}

namespace {

void showpos (std::istream & is)
{
	is.clear ();
	cerr << "At position " << is.tellg () << endl;
}

} // namespace

void z80snap::Snapshot::loadblockv1 (std::istream & is, bool debug)
{
	CHAR * ptr= memory + 0x4000;
	size_t pos= 0;
	size_t limit= 0xC000U + 1 + 100;
	// Can be at most 48K, 1 byte extra for end mark detection.
	const CHAR mark= '\xED';
	CHAR c;
	if (debug)
		cerr << "Loading V1 block" << endl;
	for (;;)
	{
		c= is.get ();
		if (! is)
		{
			showpos (is);
			throw "Error reading block";
		}
		if (c != mark)
		{
			if (pos >= limit)
				throw "Block too long";
			* (ptr++)= c;
			++pos;
		}
		else
		{
			c= is.get ();
			if (! is)
			{
				if (pos >= limit)
					throw "Block too long";
				* (ptr++)= mark;
				++pos;
				continue;
			}
			if (c != mark)
			{
				if (pos >= limit)
					throw "Block too long";
				* (ptr++)= mark;
				++pos;
				if (pos >= limit)
					throw "Block too long";
				* (ptr++)= c;
				++pos;
				continue;
			}
			CHAR times= is.get ();
			if (! is)
			{
				showpos (is);
				throw "Error reading block";
			}
			if (times == 0)
			{
				// End mark.
				if (pos > 1 && memory [0x4000 + pos - 2] != 0)
					cerr << "WARNING: non conforming "
						"end of block mark" << endl;
				--pos;
				break;
			}
			c= is.get ();
			if (! is)
			{
				showpos (is);
				throw "Error reading block";
			}
			if (debug)
				cerr << "Byte " << int (c) <<
					" repeated " << int (times) << endl;
			for ( ; times > 0; --times)
			{
				if (pos >= limit)
					throw "Block too long";
				* (ptr++)= c;
				++pos;
			}
		}
	}
	if (pos != 16 * 1024 && pos != 48 * 1024)
		cerr << "WARNING: uncompressed block is not 16K or 48K long,"
			" is " << pos << " bytes long" << endl;
}

// End of z80snap.cpp
