// z80todisc.cpp
// Revision 5-oct-2003

#include "z80snap.h"

#include <string>
#include <iostream>
#include <fstream>
#include <stdexcept>
#include <map>

const std::string progname ("z80todisc");
const std::string version ("1.0.0\n(C) 2003 Julian Albo");

class BadArgs { };

int doit (int argc, char * argv [] )
{
	using std::cout;
	using std::cerr;
	using std::endl;
	using z80snap::CHAR;

	// Analyze options.

	if (argc <= 1)
		throw BadArgs ();

	size_t autoline= 0;
	bool withoutvars= false;
	size_t endprog= 0;
	size_t addr= 0;
	char * filein= NULL;
	char * fileout= NULL;

	for (int i= 1; i < argc; ++i)
	{
		char * current= argv [i];
		if (strcmp (current, "-version") == 0)
		{
			cout << progname << " version " << version << endl;
			return 0;
		}
		else if (strcmp (current, "-auto") == 0)
		{
			if (i == argc - 1)
				throw BadArgs ();
			autoline= atoi (argv [i + 1] );
			++i;
		}
		else if (strcmp (current, "-novar") == 0)
		{
			withoutvars= true;
		}
		else if (strcmp (current, "-PROG") == 0)
		{
			if (i == argc - 1)
				throw BadArgs ();
			addr= atoi (argv [i + 1] );
			++i;
		}
		else if (strcmp (current, "-VARS") == 0)
		{
			if (i == argc - 1)
				throw BadArgs ();
			endprog= atoi (argv [i + 1] );
			++i;
		}
		else
		{
			if (filein == NULL)
			{
				filein= current;
				cerr << "Filein: " << filein << endl;
			}
			else
			{
				if (fileout == NULL)
				{
					fileout= current;
					cerr << "Fileout: " << fileout << endl;
				}
				else
					throw BadArgs ();
			}
		}
	}

	if (! filein || ! fileout)
		throw BadArgs ();

	// Load snapshot.

	std::ifstream is (filein, std::ios::in | std::ios::binary);
	if (! is.is_open () )
		throw "Can't open input file";

	z80snap::Snapshot snap (is);
	is.close ();

	// The Basic program is stored in the area delimited
	// by the system variables PROG and VARS.
	// The area of variables is from VARS to E_LINE - 1

	size_t sysvarVARS= 23627;
	size_t sysvarPROG= 23635;
	size_t sysvarE_LINE= 23641;

	if (endprog == 0)
		endprog= snap.peek16 (sysvarVARS);
	if (addr == 0)
		addr= snap.peek16 (sysvarPROG);
	if (endprog <= addr)
		throw "No Basic program";

	size_t endvar= withoutvars ? endprog :
		snap.peek16 (sysvarE_LINE) - 1;

	size_t basicsize= endprog - addr;
	size_t totalsize= endvar - addr;

	// PLUS3DOS header.

	CHAR plus3 [128] = {
		'P', 'L', 'U', 'S', '3', 'D', 'O', 'S', // Identifier.
		0x1A, // CP/M soft end of file.
		1,    // Issue number.
		0,    // Version number.
	};
	size_t filesize= totalsize + 128;
	plus3 [11]= filesize & 0xFF;
	plus3 [12]= (filesize >> 8) & 0xFF;
	plus3 [13]= (filesize >> 16) & 0xFF;
	plus3 [14]= (filesize >> 24) & 0xFF;
	//plus3 [15] is already 0 (type: Basic program).
	// Length of program + vars.
	plus3 [16]= totalsize & 0xFF;
	plus3 [17]= (totalsize >> 8) & 0xFF;
	// Autostart line (if < &8000).
	if (autoline == 0)
		autoline= 0x8000;
	plus3 [18]= autoline & 0xFF;
	plus3 [19]= (autoline >> 8) & 0xFF;
	// Length of program.
	plus3 [20]= basicsize & 0xFF;
	plus3 [21]= (basicsize >> 8) & 0xFF;
	// Compute checksum
	CHAR check= 0;
	for (int i= 0; i < 127; ++i)
		check+= plus3 [i];
	plus3 [127]= check;

	// Write the file.

	std::ofstream os (fileout, std::ios::out | std::ios::binary);
	if (! os.is_open () )
		throw "Can't open output file";
	os.write (reinterpret_cast <char *> (plus3), sizeof (plus3) );
	if (! os)
		throw "Error writing";
	os.write (reinterpret_cast <char *> (snap.basememory () + addr),
		totalsize);
	size_t round= 128 - (totalsize % 128);
	if (round != 128)
	{
		char aux [128]= {};
		os.write (aux, round);
	}

	return 0;
}

int main (int argc, char * argv [] )
{
	using std::cout;
	using std::cerr;
	using std::endl;
	using std::flush;

	try
	{
		return doit (argc, argv);
	}
	catch (BadArgs &)
	{
		cerr << "Usage: " << progname << " "
			"[ -version ] "
			"[ -auto line_num ] "
			"[ -novar ]\n\t"
			"[ -PROG mem_addr ] "
			"[ -VARS mem_addr ]\n\t"
			"file.z80 file.bas" <<
			endl;
		return 1;
	}
	catch (std::exception & e)
	{
		cout << flush;
		cerr << "ERROR: " << e.what () << endl;
		return 1;
	}
	catch (const char * str)
	{
		cout << flush;
		cerr << "ERROR: " << str << endl;
		return 1;
	}
}


// End of z80todisc.cpp
